/*!
  \file ObjCmdFormatters.h

  \brief Different SkypeObjCmdFormatter policy implementations
  
  $Id: ObjCmdFormatters.h,v 1.2 2006/04/17 06:44:14 vit Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   01.04.2006 14:25:06 MSD
*/

#ifndef OBJ_CMD_FORMATERS_H
#define OBJ_CMD_FORMATERS_H

#include "SkypeConnection.h"
#include "AttrsDesc.h"

/*!
  \defgroup SkypeObjCmdFormatter SkypeObjCmdFormatter policy of Ice::SkypeObject template   
 
  This policy is responsible for formatting GET/SET command associated with the attributes of Skype object.
  See documentation on #Ice::SkypeObject for details.
*/

namespace Ice {
/*!
  \brief SkypeObjCmdFormatter policy implementation for Skype objects with identifier.

  \param Attrs		The type of Skype object's attributes
  \param SkypeObj	The type of Skype Object.  This object should have functions "GetId()" with int return type
  			and "GetName" with const std::string& return type.  GetId function should return integer identifier
			of the object. GetName function should return its Skype name.
  
  \ingroup SkypeObjCmdFormatter			
*/  
template <typename Attrs, class SkypeObj> 
class SkypeIdObjCmdFormatter {
public:
	/*!
	  \brief Stores the references to the Skype object and SkypeConnection

	  \param _host	Reference to the Skype object which uses SkypeIdObjCmdFormatter
	  \param _sc	Connection to the Skype
	*/
	SkypeIdObjCmdFormatter(SkypeObj& _host, SkypeConnection& _sc);
	/*!
  	  \brief Formats GET (attribute value) command and sends it to the SkypeConnection (without flushing)
  
	  \param _attr	Requested attribute
	*/
	void FormatGetCmd(const Attrs& _attr);
	/*!
	  \brief Formats SET (attribute value) command and sends it to the SkypeConnection (without flushing)

	  \param _attr	The name of the setted attribute
	  \param _value The value to be set
	*/
	template <typename T> void FormatSetCmd(const Attrs& _attr, const T& _value);
private:
	/// Our host - SkypeObject 
	SkypeObj& host;
	/// Connection to the Skype
	SkypeConnection& sc;
	/// Object attributes description 
	AttrsDescription<Attrs> desc;
};	

/*!
  \brief SkypeObjCmdFormatter policy implementation for Skype objects without identifier.

  \param Attrs		The type of Skype object's attributes
  \param SkypeObj	The type of Skype Object.  This object should have function "GetName" with const std::string& return 
  			type.  GetName member function should return its Skype name.

  \ingroup SkypeObjCmdFormatter			
*/  
template <typename Attrs, class SkypeObj> 
class SkypeNoIdObjCmdFormatter {
public:
	/*!
	  \brief Stores the references to the Skype object and SkypeConnection
	  
	  \param _host	Reference to the Skype object which uses SkypeIdObjCmdFormatter
	  \param _sc	Connection to the Skype
	*/
	SkypeNoIdObjCmdFormatter(SkypeObj& _host, SkypeConnection& _sc); 
	/*!
  	  \brief Formats GET (attribute value) command and sends it to the SkypeConnection (without flushing)
  
	  \param _attr	Requested attribute
	*/
	void FormatGetCmd(const Attrs& _attr);
	/*!
	  \brief Formats SET (attribute value) command and sends it to the SkypeConnection (without flushing)

	  \param _attr	The name of the setted attribute
	  \param _value The value to be set
	*/
	template <typename T> void FormatSetCmd(const Attrs& _attr, const T& _value);
private:
	/// Connection to the Skype
	SkypeConnection& sc;
	/// Our host - SkypeObject 
	SkypeObj& host;
	/// Object attributes description 
	AttrsDescription<Attrs> desc;
};	

/**********************************************************************************************************/
/************************************ Implementation ******************************************************/
/**********************************************************************************************************/

//************************************ SkypeIdObjCmdFormatter*********************************************//

template <typename Attrs, class SkypeObj> 
inline SkypeIdObjCmdFormatter<Attrs, SkypeObj>::SkypeIdObjCmdFormatter(SkypeObj& _host, SkypeConnection& _sc)
			      : host(_host), sc(_sc)
{};

template <typename Attrs, class SkypeObj> 
void SkypeIdObjCmdFormatter<Attrs, SkypeObj>::FormatGetCmd(const Attrs& _attr)
{
  sc << "GET " << host.GetName() << " " << host.GetId() << " " << desc[_attr];	
};

template <typename Attrs, class SkypeObj> 
template <typename T> void SkypeIdObjCmdFormatter<Attrs, SkypeObj>::FormatSetCmd(const Attrs& _attr, const T& _value)
{
  sc << "SET " << host.GetName() << " " << host.GetId() << " " << desc[_attr] << " " << _value;	
};	

//************************************ SkypeNoIdObjCmdFormatter*******************************************//

template <typename Attrs, class SkypeObj> 
SkypeNoIdObjCmdFormatter<Attrs, SkypeObj>::SkypeNoIdObjCmdFormatter(SkypeObj& _host, SkypeConnection& _sc)
					  : sc(_sc), host(_host)
{};

template <typename Attrs, class SkypeObj> 
void SkypeNoIdObjCmdFormatter<Attrs, SkypeObj>::FormatGetCmd(const Attrs& _attr)
{
  sc << "GET " << host.GetName() << " " << desc[_attr];	
};

template <typename Attrs, class SkypeObj> 
template <typename T> void SkypeNoIdObjCmdFormatter<Attrs, SkypeObj>::FormatSetCmd(const Attrs& _attr, const T& _value)
{
  sc << "SET " << host.GetName() << " " << desc[_attr] << " " << _value;	
};

}; // namespace Ice
#endif // OBJ_CMD_FORMATERS_H

