/*!
  \file Search.h

  \brief Search-related stuff

  $Id: Search.h,v 1.14 2006/06/15 09:43:29 tarle Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   25.02.2006 20:36:11 MSD
*/

#ifndef SK_SEARCH_H
#define SK_SEARCH_H

#include "SkypeConnection.h"
#include "AttrsDesc.h"

#include <vector>
#include <set>
#include <algorithm>

#include <boost/tokenizer.hpp>

namespace Ice {

/*!
  \brief Dummy implementation of the Search policy of the SkypeObjManager template

  	 This Search policy implementation is used in SkypeObManagers which don't support 
	 search interface.
	 
  \param Host		SkypeObjManager instance
  \param SearchTypes	Doesn't matter.
*/  
template < class Host, class SearchTypes = Loki::NullType>
class NoSearch {
public:
	/*!
	  \brief  NoSearch constructor.  Does nothing.	
  
	  \param _host	Reference to SkypeObjManager instance
	  \param _sc	Reference to SkypeConnection
	  \param  _lf	Logger to use by NoSearch's functions

	*/
	NoSearch(Host& _host, SkypeConnection& _sc, const LogFunc_t& _lf = DummyLogger) {};	
};


/*!
  \brief Search interface for SkypeObjManager template.  It implements Search policy.

  There are two parameters of this template - Host and SearchTypes.  Host is the (child of the) instance of the SkypeObjManager.
  SearchTypes parameter represents typese of possible search requests (Example for call objects: search for latest calls, 
  search for missed calls and so on). This is enumeration type.  The specialization of AttrsDescription template should exist for 
  this (SearchTypes) type.

  The following types should be typedef'ed inside that specialization: ObjType_t (type of objects controlled by SkypeObjManager and 
  returned as search results) and ObjAttrs_t (type of object's attributes). String representation of the SearchTypes should be stored in
  the AttrsDescription template specialization for SearchTypes type.  It should be accessable via operator[] of the AttrsDescription specialization.

  To start search call Search method of this template.  

  \param Host		SkypeObjManager instance
  \param SearchTypes	Types of the search.  See above. 
*/  
template <class Host, typename SearchTypes>
class Searcher : virtual public LogAbility { 
public:
	/*!
	  \brief  Searcher initialization

	  \param _host	Reference to the host object
	  \param _sc	Reference to Skype Connection  
	  \param  _lf   Reference to the logger. This logger will be used to log any messages inside the object.
	*/
	Searcher(Host& _host, SkypeConnection& _sc, const LogFunc_t& _lf = DummyLogger);
	/// Description of all available search types	
	typedef AttrsDescription<SearchTypes> SearchTypesDesc_t;
	/// Set of object's (which are the results of the search procedure) attributes
	typedef std::set<typename SearchTypesDesc_t::ObjAttrs_t> AttrSet_t;
	/// Type of object - elements of the search result
	typedef typename SearchTypesDesc_t::ObjType_t ObjType_t;
	/*! 
	  \brief Functor of this type is called to report the search results to the initiator of the search.
	  
	  This only argument of this functor is the vector of pointers to the objects.  These objects are the result of the search
	  procedure. 

	  The functor returns bool value.  If it returns true, all objects are stored inside the object manager.  Otherwise, they
	  are released.  If you are going to work with the search results, return true. Otherwise, return false.
	*/
	typedef Loki::Functor<bool, LOKI_TYPELIST_1(const std::vector<ObjType_t*>&)> SearchResult_t;
	/*!
	  \brief  Start the search procedure

	  This procedure initiates the search.  It is not allowed to call this method while previous search is in progress.
	  Upon search completion, _cbk is called.  The return value of _cbk is analyzed.  See #SearchResult_t description fr details.

	  \param  _what		The type of the search 
	  \param  _cbk		Callback functor to report the search results.  See #SearchResult_t description for details
	  \param  _attrs	Object attributes which are required to be set in the search results (skype objects)
	  \param  _param	Search parameter
	*/
	void Search(const SearchTypes& _what, const SearchResult_t& _cbk, const AttrSet_t& _attrs, const std::string& _param = "");
private:
	/*!
	  \brief Search query response handler
  
	  	 Handles the skype's response to the search query.
		 Gets all object identifiers from the search resul, checks their existence in the host (Skype object manager).
		 Requests all the required object's attributes.  Registers #FullObjWasCreated and #ObjWasChanged methods in the 
		 host (Skype object manager)
		 
	  \param _resp Skype's response
	*/
	void ProcessResp(const std::string& _resp);
	/// This method is called when object is created by the host (skype object manager)  
	/// This object can be created as a search result or independently
	bool ObjWasCreated(ObjType_t& _obj);
	/// This method is called when object with all the required attributes is created by the host (skype object manager) 
	bool FullObjWasCreated(ObjType_t& _obj);
	/// This method is called when one of the required object attributes was set
	void ObjWasChanged(ObjType_t& _obj, typename SearchTypesDesc_t::ObjAttrs_t _attr, const std::string&);
	/// Throws SkypeInvalidSearch exception
	void ErrHandler(const int& _code, const std::string& _err_msg, const std::string& _full_err_msg);
	/// Called when finish was searched.  This method executes user callback-function supplied as the _cbk argument of the #Search method 
	bool FinishSearch(ObjType_t* _obj = 0);
	/// Set to #ProcessResp in the constructor 
	const SimpleHandler_t resp_h;
	/// Set to #ErrHandler in the constructor
	const ErrorHandler_t err_h;
	/// Identifier of the #ObjWasCreated listener
	unsigned obj_listener_id;
	/// Identifiers of the #FullObjWasCreated listener
	unsigned full_listener_id;
	/// Sets to true by the #Search method.  Set to false upon search completion
	bool search_in_progress;
	/// Reference to the host object (skype object manager)
	Host& host;
	/// Reference to the Skype connection
	SkypeConnection& sc;
	/// Callback functor which is executed after the completing of the search procedure
	SearchResult_t search_cbk;
	/// Describes all available search query types 
	const SearchTypesDesc_t SearchStrings;
	/// Objects created while search is in progress (search_in_progress == true)
	std::vector<ObjType_t*> created_objs;
	/// Object's identifier type
	typedef typename ObjType_t::ID_t ObjId_t;
	/// Identifiers of the searhed objects.  Filled by the #ProcessResp method 
	std::vector<ObjId_t> searched_ids;
	/// Object attributes which are required to be set in the search results
	AttrSet_t required_attrs;
	/// Results of the search procedure
	std::vector<ObjType_t*> search_results;
	/// Identifiers of active watchers
	std::map<ObjId_t, unsigned> active_watchers;
};


/**********************************************************************************************************/
/************************************ Implementation ******************************************************/
/**********************************************************************************************************/

template <class Host, typename SearchTypes>
Searcher<Host, SearchTypes>::Searcher(Host& _host, SkypeConnection& _sc, const LogFunc_t& _lf)
			    : LogAbility(_lf),
			      resp_h(this, &Searcher<Host, SearchTypes>::ProcessResp), 
			      err_h(this, &Searcher<Host, SearchTypes>::ErrHandler),
			      obj_listener_id(0), full_listener_id(0), search_in_progress(0),
			      host(_host), sc(_sc)
{};

template <class Host, typename SearchTypes>
void Searcher<Host, SearchTypes>::Search(const SearchTypes& _what, const SearchResult_t& _cbk, const AttrSet_t& _attrs, const std::string& _param)
{
  Assert(search_in_progress == false);
  Assert(obj_listener_id == 0);
  Assert(full_listener_id == 0);
  AttrSet_t empty;  
  typename Host::ObjCreated_t cbk(this, &Searcher<Host, SearchTypes>::ObjWasCreated);
  obj_listener_id = host.RegisterListener(empty, cbk);  
   
  sc << "SEARCH " << SearchStrings[_what];
  if(_param.length()) sc << " " << _param;
 
  sc << Flush(resp_h, err_h, 0);
  search_in_progress = true;
  required_attrs = _attrs;
  search_cbk = _cbk;
};	

template <class Host, typename SearchTypes>
void Searcher<Host, SearchTypes>::ProcessResp(const std::string& _resp)
{
  std::string::size_type i = _resp.find(' ');
  Assert(i != 0);
  Assert(i != std::string::npos);
  i++;

  boost::char_separator<char> sep(" ,");
  
  typedef boost::tokenizer < boost::char_separator<char> > Tokenizer_t;
  typedef Tokenizer_t::iterator TokIt_t;

  const std::string& parsed_str = _resp.substr(i);
  Tokenizer_t toks(parsed_str, sep);
  TokIt_t p = toks.begin();
  
  typename Host::ObjCreated_t cbk(this, &Searcher<Host, SearchTypes>::FullObjWasCreated);
  full_listener_id = host.RegisterListener(required_attrs, cbk);

  Assert(obj_listener_id != 0); host.UnRegisterListener(obj_listener_id); obj_listener_id = 0;
  typedef typename std::vector<ObjType_t*>::iterator IT;

  for(TokIt_t p = toks.begin();p != toks.end();++p) {
	const ObjId_t& id = GetFromString<ObjId_t>(*p); 
	host.UnReleaseObj(id);
	bool should_add_to_search_ids = true;
	// Is there this object in created_objs?
	ObjType_t* obj = 0;
	for(IT p=created_objs.begin();p!=created_objs.end();++p) {
		if((*p)->GetId() == id) {
			obj = *p;
			created_objs.erase(p);
			break;
		};	
	};	
	// Is there this object in ObjManager already?
	if(!obj) obj = host.find(id, true);
	if(obj == 0) host.Create(id);
	else {
		typedef typename AttrSet_t::const_iterator CI;
		typename SearchTypesDesc_t::ObjAttrs_t last_attr;
		bool should_watch(false); 
		for(CI p=required_attrs.begin();p!=required_attrs.end();++p) {
			if(obj->find(*p)==obj->end()) {
				obj->AskAttr(*p);
				should_watch = true;
				last_attr = *p;
			};	
		};
		if(should_watch) {
			typedef typename ObjType_t::Watchers_t Watchers_t;	
			Watchers_t& w = obj->GetWatchers();
			AttrSet_t attrs;
			attrs.insert(last_attr);
			typename Watchers_t::ObjChanged_t cbk(this, &Searcher<Host, SearchTypes>::ObjWasChanged);
//			typename Watchers_t::WatcherDetailsPtr_t wd = new typename Watchers_t::WatcherDetails(attrs, cbk);
			Assert(active_watchers.find(obj->GetId()) == active_watchers.end());
			active_watchers[obj->GetId()] = w.RegisterWatcher(attrs, cbk);
		} else {
			should_add_to_search_ids = false;
			search_results.push_back(obj);
		};	
	};
	if(should_add_to_search_ids) searched_ids.push_back(id);
  };
  
  for(IT p=created_objs.begin();p!=created_objs.end();++p) host.ReleaseObj(*(*p));
  created_objs.clear();
	
  if(searched_ids.empty()) {
  	FinishSearch();
  };
};

template <class Host, typename SearchTypes>
bool Searcher<Host, SearchTypes>::ObjWasCreated(ObjType_t& _obj)
{
  Assert(search_in_progress == true);	
  
  created_objs.push_back(&_obj);
  return search_in_progress;
};

template <class Host, typename SearchTypes>
bool Searcher<Host, SearchTypes>::FinishSearch(ObjType_t* _obj)
{
  bool r = search_cbk(search_results);
  search_in_progress = false;
  
  Assert(full_listener_id); host.UnRegisterListener(full_listener_id); full_listener_id = 0;
  
  if(!r) {
	typedef typename std::vector<ObjType_t*>::iterator IT;
	if(_obj) {
		IT p = std::find(search_results.begin(), search_results.end(), _obj);
		Assert(p!=search_results.end());
		search_results.erase(p);
	};	
	for(IT p=search_results.begin();p!=search_results.end();++p) host.ReleaseObj(*(*p));
  };	
  search_results.clear();
  
  return r;
};

template <class Host, typename SearchTypes>
bool Searcher<Host, SearchTypes>::FullObjWasCreated(ObjType_t& _obj)
{
  Assert(search_in_progress == true)	
  
  typedef typename std::vector<ObjId_t>::iterator IT;
  IT p = std::find(searched_ids.begin(), searched_ids.end(), _obj.GetId());
  
  bool r = (p != searched_ids.end());
  if(r) {
	searched_ids.erase(p);
	search_results.push_back(&_obj);
  };
  
  if(searched_ids.empty()) r = FinishSearch(&_obj);
  
  return r;  
};

template <class Host, typename SearchTypes>
void Searcher<Host, SearchTypes>::ObjWasChanged(ObjType_t& _obj, typename SearchTypesDesc_t::ObjAttrs_t _attr, const std::string&)
{
  typedef typename AttrSet_t::const_iterator CI;
  for(CI p=required_attrs.begin();p!=required_attrs.end();++p) Assert(_obj.find(*p) != _obj.end());
  
  typedef typename std::map<ObjId_t, unsigned>::iterator IT;
  ObjId_t id = _obj.GetId();
  IT p = active_watchers.find(id);
  Assert(p != active_watchers.end());
  _obj.GetWatchers().UnRegisterWatcher(p->second);
  active_watchers.erase(p);
  FullObjWasCreated(_obj); 
};

template <class Host, typename SearchTypes>
void Searcher<Host, SearchTypes>::ErrHandler(const int& _code, const std::string& _err_msg, const std::string& _full_err_msg)
{
  throw SkypeInvalidSearch(typeid(*this).name(), _code, _err_msg, _full_err_msg);	
};	

}; // namespace Ice

#endif // SK_SEARCH_H
