/*!
  \file SkypeConnection.h

  \brief Skype connection module
  
  $Id: SkypeConnection.h,v 1.15 2006/05/26 07:13:58 vit Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   27.07.2005 11:26:34 MSD
*/

#ifndef SKYPE_CONNECTION_H
#define SKYPE_CONNECTION_H

#include "LowLevel.h"
#include "utils.h"

#include <map>
#include <boost/regex.hpp>

/// Simple handler.  Can be used as events or command response handler.  See documentation on SkypeConnection for details
typedef Loki::Functor<void, LOKI_TYPELIST_1(const std::string&)> SimpleHandler_t;

/// SkypeConnection event with regexp data.  See documentation on SkypeConnection for details
/// Note: match[0] always contains full matched string (original event received from Skype)
typedef Loki::Functor<void, LOKI_TYPELIST_1(const boost::cmatch&)> RegExNotify_t;

/*!
  \brief Skype errors (API commands level, not Skype <-> application transport layer) handler

   Parameters (from left to right):
	- Integer skype error code 
	- Human-readable error message generated in accordance with Skype error code 
	- Full Skype error message received from Skype 

   This type is used as second parameter of Flush manipulator.  
   See documentation on SkypeConnection::operator<<(const Flush& _f) for details.	
*/  
typedef Loki::Functor<void, LOKI_TYPELIST_3(const int&, const std::string&, const std::string&)> ErrorHandler_t;

/*! This structure is used to implement Flush manupulator with objects of SkypeConnection class

  Flush manipulator takes 3 parameter: response handler, error handler and event sending mode.
  If last parameter is set to true (the default), events are sent to all registered event handlers
  (see #SkypeConnection::RegisterCallback) upon receiving response from Skype.  They are sent only 
  after the execution of the response handler.  If the event sending mode is set to false, only 
  response handler is called.  In case the last parameter is omitted, it is assumed to be true.
*/	
struct Flush {
	/// Stores references to response and error handlers and event sending mode 
	Flush(const SimpleHandler_t& _sh, const ErrorHandler_t& _eh, bool _send_event = true);
	/// This constructor is required to be able to store objects of this class inside std::map
	Flush() {};
	/// Response handler
	SimpleHandler_t sh;
	/// Error handler
	ErrorHandler_t eh;
	/// Should we notify about this response via event mechanism also?
	bool send_event;
};

/*!
  \brief Connection with Skype

  This class is responsible for Skype connection establishment and maintenance. There are two kinds of messages:
  Skype -> application and application -> Skype (called commands).  There are two kinds of Skype-> application 
  messages: notifications (also called events) and responses to commands. If #DisableIds mode is turned on, there is
  no way to distinguish command responses from events.  
  
  Commands are sent to Skype with the help of operator<< of this class (defined in it's base classes, actually). 
  There is a new #Flush method is introduced in this class.  If you want to distinguish events and command responses
  you should use this method for sending commands to Skype.  Skype can respond with some data related to command
  being sent or with error. In case of successfull command execution SkypeConnection will call first #Flush parameter.
  Otherwise second #Flush parameter will be called.  Command responses are processed as responses only if you sent
  your commands via #Flush manipulator, not #flush!  Also see documentation on SkypeConnection::operator<<(const Flush& _f).
  
  Notifications are processed by user-defined call-back functors of types #RegExNotify_t and #SimpleHandler_t. 
  These functors should be registered via #RegisterCallback method call and unregistered via call to #UnRegisterCallback.
  operator() of this class gets the new Skype events and matches received event with all registered regular expressions
  (The order of comparisons with different registered regular expressions is unspecified).  If match is found,  
  associated call-back functor(s) is (are) called.  If no matched regular expression is found, default event handler is called.
  Reference to the default handler should be given to SkypeConnection via its constructor. 

  There are two functor types suitable for using as call-back functors - #RegExNotify_t and #SimpleHandler_t.  First of them
  should be used if your regular expression (registrated via #RegisterCallback) contains sub-expressions and you want to
  utilize their values during event processing. There are no more differences in these functor types. 
*/ 
class SkypeConnection : public RawSkypeConnection<> {
public:
	/*!
  	  \brief Stores all data required for interaction with Skype

  	  \param _app_name	Application's name (usually argv[0]).  
  	  \param _def		Default handler of Skype events.  See documentation on SkypeConnection for details.
  	  \param _log_func	Logger to use by SkypeConnection's functions
	*/
	SkypeConnection(const std::string& _app_name, const SimpleHandler_t& _def, const LogFunc_t& _log_func = DummyLogger);
	/*!
  	  \brief Skype event handler (simple) registration.  See documentation on SkypeConnection for
		 additional information.

  	  \param _regex	Regular expression
  	  \param _f	Event handler.  It will be executed if received skype event matches _regex
	  
	  \throw AssertionViolation (Debug mode only) If _regex can not be compiled. 
	  
	  \return Identifier of the registered event handler
	*/
	unsigned RegisterCallback(const std::string& _regex, const SimpleHandler_t&  _f);
	/*!
  	  \brief Skype event handler (with regexp data) registration. See documentation on SkypeConnection for
	  	 additional information.
  
  	  \param _regex	Regular expression
  	  \param _f	Event handler.  It will be executed if received skype event matches _regex
	
	  \throw AssertionViolation (Debug mode only)  If _regex can not be compiled. 

	  \return Identifier of the registered event handler
	*/
	unsigned RegisterCallback(const std::string& _regex, const RegExNotify_t& _f);
	/*!
	  \brief Unregistrates Skype event handler previously registered with RegisterCallBack functions.
  
		 If there is no registered event handler with id equal to _id for regular expression _regex, nothing happens.

	  \param _regex Regular expression associated with the registered handler 	 
	  \param _id    Skype event handler identifier
	*/
	void UnRegisterCallback(const std::string& _regex, const unsigned& _id);
	/*!
  	  \brief Sends command to Skype with registration.  Stores response and error handlers for later usage. 

	  Should be called only if #IsIdsEnabled returns true.  In other cases, flush() should be used.    
	  See documentation on Flush for details.
	  
	  Typical usage of this operator is shown below (it is assumed that type of sc variable is SkypeConnection):  
	  
 \code 
 using std::string;
 // Response handler 
 void RespH(const string& s);
 // Error handler
 void ErrH(const int& r, const string& e, const string& m); 
 
 // don't send events 
 sc << "PING" << Flush(RespH, ErrH, 0);

 // send events
 sc << "PING" << Flush(RespH, ErrH, 1);
 // send events	
 sc << "PING" << Flush(RespH, ErrH); 

 \endcode
	  	
	  \throw AssertionViolation (Debug mode only) If IsIdsEnabled returns false. 

  	  \param _f Flush object (usually temporary).  See typical usage example above. 
	*/
	SkypeConnection& operator<<(const Flush& _f);
	/// Calls related #SkypeCmdFormatter operators "<<" 
	template<class T> SkypeConnection& operator<<(T _arg);
	/// Brief parent type name.  Only for additional comfort
	typedef SkypeCmdFormatter< RawSkypeConnection<> > CmdFormatter_t;
	/// "<< flush", "<< EnableIds", "<< DisableIds" syntax. Calls related #SkypeCmdFormatter operator "<<" 
	SkypeConnection& operator<<(CmdFormatter_t& (*_f)(CmdFormatter_t&));
private:
	SkypeConnection(const SkypeConnection&);
	SkypeConnection& operator=(const SkypeConnection&);
	/*!
  	  \brief Main skype events handler.  It will be executed upon Skype event receiving by RawSkypeConnection
	         implementation.
  
  	  \param _msg Event received from Skype.
	*/
	void Dispatcher(const std::string& _msg);
	///  Default Skype events handler.  It will be called if no regular expression matches are found
	SimpleHandler_t default_cbk;
	/// Regular expression <-> linked call-back functor structure
	typedef struct {
		/// Simple notify functor(s). 
		Registrator<SimpleHandler_t, unsigned short> simple_f;
		/// RegExp notify functor(s). 
		Registrator<RegExNotify_t, unsigned short> regex_f;
		/// Compiled regular expression
		boost::regex regex;
	} RegExFuncPair_t;	
	/// Type for storing functors <-> regular expressions relations
	typedef std::map<std::string, RegExFuncPair_t> FuncLst_t; 
	/// Key of this array is textual representation of regular expression
	FuncLst_t regex_map;
	/// Type for storing active (currently executed by skype and non-responsed) commands
	typedef std::map<int, Flush> ActiveCmds_t;
	/// Active commands
	ActiveCmds_t active_cmds;
	/// Regular expression to parse error response received from Skype
	const boost::regex error_parser;
	/*!
	  \brief Sends event to all registered event handlers.

	  \param _event Event received from the Skype
	*/
	void SendEvent(const std::string& _event);
	/*!
  	  \brief Looks for _regex in regex_map. If not found, compiles _regex and creates new regex_map element   
  
  	  \param _regex Regular expression passed to RegisterCallback
	  \throw AssertionViolation If _regex can not be compiled (invalid regular expression) 
	  
  	  \return iterator pointed to regex_map element
	*/
	FuncLst_t::iterator CreateRegExFuncPair(const std::string& _regex);
	/*!
	  \brief This template is used to call callback functors (event handlers) associated with regular expressions	
	
	  \param Functor Type of callback functor: SimpleHandler_t or RegExNotify_t
	*/
	template <class Functor> 
	class RunCallBacks {
	public:
		/*!
		  \brief Stores the argument of callback functor

		  \param _arg Functor's argument
		*/
		RunCallBacks(typename Functor::Parm1 _arg);
		/*!
	  	  \brief Calls functor _f with argument stored in the constructor

		  This method is called by Registrator::for_each function.
		  
		  \param _id  Identifier of the call back functor
		  \param _f   Call back functor
		*/
		void operator()(const unsigned short& _id, Functor& _f) const;
	private:
		/// Argument of the callback functor
		typename Functor::Parm1 arg;
	};	
};	

/**********************************************************************************************************/
/************************************ Implementation ******************************************************/
/**********************************************************************************************************/

template <class Functor> 
inline SkypeConnection::RunCallBacks<Functor>::RunCallBacks(typename Functor::Parm1 _arg)
			: arg(_arg) 
{};

template <class Functor> 
inline void SkypeConnection::RunCallBacks<Functor>::operator()(const unsigned short& _id, Functor& _f) const
{ _f(arg); };

template<class T> 
inline SkypeConnection& SkypeConnection::operator<<(T _arg)
{
  CmdFormatter_t& me = *this;
  me << _arg;  	

  return *this;
};	

inline SkypeConnection& SkypeConnection::operator<<(SkypeConnection::CmdFormatter_t& (*_f)(SkypeConnection::CmdFormatter_t&))
{
  CmdFormatter_t& me = *this;
  me << _f;
  
  return *this;
};

#endif // SKYPE_CONNECTION_H
