/*!
  \file unix/LowLevel-dbus.h

  \brief Low-level ++Skype API implementation for unix OS. D-BUS variant

  $Id $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   18.10.2006 08:19:42 MSD
*/

#ifndef UNIX_LOWLEVEL_DBUS_H
#define UNIX_LOWLEVEL_DBUS_H

#include "../Cfg.h"

#ifdef USE_DBUS_INTERFACE 

#include "../SkypeExceptions.h"
#include "../utils.h"

/// This define is required to be set by the D-BUS library used to connect to Skype 
#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus.h>

extern "C" {
#include <sys/select.h>
#include <sys/types.h>
#include <unistd.h>	
};

#include <list>

/// Implementation of RawSkypeConnection interface for the linux DBUS API
typedef RawSkypeConnection<os_nix> UnixDbusRawSkypeConnection_t;

/// Implementation of RawSkypeConnection interface for the unix (currently linux) API
typedef UnixDbusRawSkypeConnection_t UnixRawSkypeConnection_t;

/// RawSkypeConnection specialization for unix OS
template <> class RawSkypeConnection<os_nix> : public SkypeCmdFormatter< RawSkypeConnection<os_nix> >,
					       public LogAbility 
{
public:
	/*!
	  \brief Stores functor which will be executed upon messages (Skype -> application) receiving

	  \param _cbk Call-back functor. It will be executed for any message received from Skype
	  \param _app_name Application's name (usually argv[0]).  
	  \param _log_func Logger to use inside RawSkypeConnection 
	*/  
	RawSkypeConnection(const RawNotify_t& _cbk, const std::string& _app_name, LogFunc_t _log_func = DummyLogger);	
	/// D-BUS's bus used for communication with skype. SessionBus means session (per-user) bus.
	/// SystemBus means operation system-wide bus. 
	typedef enum { SessionBus, SystemBus } DBusType_t;
	/*!
  	  \brief		Attach to Skype. Do nothing if already attached. 
	  
	  \param _bus_tp  	Bus to attach to.
	  \param _dont_select 	To check D-BUS bus for Skype notifications unix select is used.  It can
	  			be executed inside () operator of RawSkypeConnection<os_nix> (first policy)
				or in another piece of your code (second policy). If _dont_select is true
				then second policy is active.  And if it is false, first policy is active.
				Note, you need file descriptor for the select's readfs to call selects
				yourself.  It can be obtained via #GetBusFD method of this class.
	  \param _timeout 	Select's timeout value in microseconds. The value of this parameter
	  			doesn't used anywhere if _dont_select is set to true. 

	  \throw SkypeDisconnect Error occured during connect.		
	  \throw SkypeAccessDenied Skype user doesn't allow this application to connect to Skype
	*/
	void Attach(DBusType_t _bus_tp = SessionBus, bool _dont_select = false, unsigned _timeout = 500);
	/// Returns true if attached to Skype, false - otherwise
	bool IsAttached(void) { return attach_result; };
	/// Detach from Skype. Do nothing if already detached
	void Detach();
	/*!
	  \brief Send mesage to Skype via D-BUS. 

	  \param _cmd Skype command

	  \throw SkypeDisconnect Disconnected from Skype 
	*/  
	void flush(const std::string& _cmd);
	/*!
	  \brief Send mesage to Skype via D-BUS (calls SkypeCmdFormatter::flush())

	  \throw SkypeDisconnect Disconnected from Skype 
	*/  
	void flush();

	/*!
	  \brief	Read (if any) and process messages from Skype (Skype notifications and 
	  		command responses).

	  		This is a very simple method. It checks for new messages in queue and
			runs user-defined call-back functor (see RawSkypeConnection constructor) 
			for each received message. See also Attach docs for some implementation 
			details.
			
	  \throw AssertionViolation  If method is executed when connection to D-BUS bus is not established.
	  			     It is thrown only if debug mode is on (DbgMode is set to debug 
				     in Cfg.h).
			
	  \return	True if at least 1 message was received.  False otherwise.
	*/  
	bool operator()();
	/*!
	  \brief Get file descriptor associated with D-BUS bus

	  \throw AssertionViolation  If method is executed when connection to D-BUS bus is not established.
	  			     It is thrown only if debug mode is on (DbgMode is set to debug 
				     in Cfg.h).
	  
	  \return D-BUS bus file descriptor. You need it if you want to call select yorself. See 
		  #Attach method docs for details.
	*/  
	int GetBusFD() const;
private:
	/// Checks dbus_error for errors and throws #SkypeDisconnect if error is found
	void CheckError();
	/// Set dbus_iter to the first parameter of _msg.  #AssertionViolation is thrown if _msg has no params
	void InitDbusIter(DBusMessage& _msg);
	/// Get the string parameter by iterator.  #AssertionViolation is thrown in case of type mismatch
	const char* GetArgStr(DBusMessageIter& _i);
	/// Get the integer parameter by iterator.  #AssertionViolation is thrown in case of type mismatch
	int GetArgInt32(DBusMessageIter& _i);
	/// Get the boolean parameter by iterator.  #AssertionViolation is thrown in case of type mismatch
	bool GetArgBool(DBusMessageIter& _i);
	/// Send msg with block.  Returns reply. Used only in #Attach method. Unrefs _msg and check for reply present via Assert
	DBusMessage* SendMsgWBlock(DBusMessage& _msg);
	/*!
	  \brief Incoming skype messages processor

	  	 Receives message(s) from skype, extracts 1st string message argument and runs cbk with this argument.
	  
	  \param _user_data Pointer to this object (this-pointer)	 
	*/  
	static DBusHandlerResult dbus_msg_handler(DBusConnection * _connection, DBusMessage * _msg, void *_user_data);
	/// Dummy handler.  Should never be executed 
	static void dbus_unregister_handler(DBusConnection*, void*);
	/// Add watch.  Executed by D-BUS library
	static dbus_bool_t dbus_add_watch(DBusWatch* _watch, void* _data);
	/// Dummy handler.  Should never be executed 
	static void dbus_remove_watch(DBusWatch*, void*);
	/// Add std::string argument to D-BUS message. Can throw AssertionViolation exception	
	void dbus_msg_add_arg(DBusMessage* _m, const std::string& _arg);
	/// Add integer argument to D-BUS message. Can throw AssertionViolation exception
	void dbus_msg_add_arg(DBusMessage* _m, const int& _arg);
	/// Call-back functor.  
	RawNotify_t cbk;
	/// Application's name
	const std::string app_name;
	/// Shouldn't we execute select inside () operator?
	bool dont_select;
	/// The value of timeout for select (in microseconds) if dont_select is false
	unsigned timeout;
	/// D-BUS replies received in flush.  They will be processed in operator()
	std::list<DBusMessage*> dbus_replies;
	/// D-BUS connection handle
	DBusConnection *dbus_cn;
	/// D-BUS watch object (used for calls to select) 
	DBusWatch *dbus_watch;
	/// Stores D-BUS error information
	DBusError dbus_error;
	/// D-BUS message parameters (arguments) iterator 
	DBusMessageIter dbus_iter;
	/// Is connection to D-BUS established?
	bool attach_result;
	/// File descriptor set, checked by select
	fd_set read_fds;
	/// File descriptor watched by select
	int dbus_fd;
	/// Application's PID.  It is required for valid answer to PING command from Skype
	pid_t pid;
};

/**********************************************************************************************************/
/************************************ Implementation ******************************************************/
/**********************************************************************************************************/

inline int UnixDbusRawSkypeConnection_t::GetBusFD() const
{
  Assert(attach_result);
  Assert(dbus_watch);

  return dbus_fd;
};

inline void UnixDbusRawSkypeConnection_t::CheckError()
{
  if(dbus_error_is_set(&dbus_error)) {
	attach_result = false;
	throw SkypeDisconnect(dbus_error.message);  
  };
};

inline void UnixDbusRawSkypeConnection_t::InitDbusIter(DBusMessage& _msg)
{
  dbus_bool_t r = dbus_message_iter_init(&_msg, &dbus_iter);	
  Assert(r == TRUE);
};

inline const char* UnixDbusRawSkypeConnection_t::GetArgStr(DBusMessageIter& _i)
{
  Assert(dbus_message_iter_get_arg_type(&_i) == DBUS_TYPE_STRING);
  return dbus_message_iter_get_string(&_i);
};

inline int UnixDbusRawSkypeConnection_t::GetArgInt32(DBusMessageIter& _i)
{
  Assert(dbus_message_iter_get_arg_type(&_i) == DBUS_TYPE_INT32);
  return dbus_message_iter_get_int32(&_i);
};

inline bool UnixDbusRawSkypeConnection_t::GetArgBool(DBusMessageIter& _i)
{
  Assert(dbus_message_iter_get_arg_type(&_i) == DBUS_TYPE_BOOLEAN);
  return dbus_message_iter_get_boolean(&_i);
};

inline DBusMessage* UnixDbusRawSkypeConnection_t::SendMsgWBlock(DBusMessage& _msg)
{
  dbus_message_set_auto_activation(&_msg, true);
  DBusMessage* reply = dbus_connection_send_with_reply_and_block(dbus_cn, &_msg, -1, &dbus_error);
  CheckError();  // Memory leak is possible (msg, reply), but is it important in this place ? :-) 
  dbus_message_unref(&_msg);
  Assert(reply != NULL);
  return reply;
};

inline void UnixDbusRawSkypeConnection_t::dbus_unregister_handler(DBusConnection*, void *)
{ Assert(0); };

inline void UnixDbusRawSkypeConnection_t::dbus_remove_watch(DBusWatch*, void*)
{ Assert(0); };

inline void UnixDbusRawSkypeConnection_t::dbus_msg_add_arg(DBusMessage* _m, const std::string& _arg)
{
  dbus_bool_t r = dbus_message_append_args(_m, DBUS_TYPE_STRING, _arg.c_str(), 0);
  Assert(r==TRUE);	
};

inline void UnixDbusRawSkypeConnection_t::dbus_msg_add_arg(DBusMessage* _m, const int& _arg)
{
  dbus_bool_t r = dbus_message_append_args(_m, DBUS_TYPE_INT32, _arg, DBUS_TYPE_INVALID);
  Assert(r==TRUE);	
};

#endif // USE_DBUS_INTERFACE 
#endif // UNIX_LOWLEVEL_DBUS_H

