/*!
  \file audio.cpp

  \brief Check audio settings.  This test correctly works only under Windows because Skype for linux doesn't support tested features 

  $Id $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   29.03.2006 17:33:49 MSD
*/

#include <boost/test/unit_test.hpp>
#include <boost/test/unit_test_suite.hpp>

#include <algorithm>
#include <typeinfo>
#include <list>

#include "h/Audio.h"
#include "h/utils.h"
#include "h/Cfg.h"

#include "SkypeTest.h"

using namespace boost::unit_test;
using namespace Ice;

SkypeConnection* SC;

void DefaultCallback(const std::string& s)
{
  fprintf(stdout, "%s: '%s'\n", __FUNCTION__, s.c_str());	
};

template <typename Attrs>
void PrintValue(const std::pair<Attrs, std::string>& _msg)
{
  fprintf(stdout, "%s/%s: Atribute %d value is '%s'\n", __FUNCTION__, typeid(_msg.first).name(), _msg.first, _msg.second.c_str());	
};

template<class Test, class TestAttr>
class CallBack {
public:
	void Do(Test&, const TestAttr& _tst_attr, const std::string& _tst_v)
	{ tst_v = _tst_v;  tst_attr = _tst_attr; };	
	std::string tst_v;
	TestAttr tst_attr;
};
 
template<class Test, class TestAttr>
void WatchTest(Test& test, TestAttr attr)
{
  typedef StateWatchers<Test, TestAttr> Watchers_t;
  Watchers_t w;
  typename Watchers_t::AttrSet_t attr_set;
  CallBack<Test, TestAttr> test_cbk;
  typename Watchers_t::ObjChanged_t cbk(&test_cbk, &CallBack<Test, TestAttr>::Do);
  unsigned ids[5];

  for(int i=0;i<5;i++) ids[i] = w.RegisterWatcher(attr_set, cbk);
  w(test, attr, "test");
  BOOST_CHECK(test_cbk.tst_v.length() == 0);
  
  for(int i=0;i<2;i++) w.UnRegisterWatcher(ids[i]);
  BOOST_CHECK_NO_THROW(w.UnRegisterWatcher(ids[0]));
  BOOST_CHECK_NO_THROW(w.clear());
  
  w.RegisterWatcher(attr_set, cbk, true);
  w(test, attr, "test");
  BOOST_CHECK(test_cbk.tst_v == "test");
  BOOST_CHECK(test_cbk.tst_attr == attr);
};

void WatchTestAudio()
{
  AudioSettings au(*SC);	
  WatchTest<AudioSettings, AudioAttrs_t>(au, auMute);	
};

template<class Test, class TestAttr, class Result> 
void DoTest(const std::list<TestAttr>& _attrs)
{
  SkypeConnection& sc = *SC;
  Result r;
  const boost::posix_time::seconds tm(1);
 
  sc << DisableIds << "PROTOCOL 99" << flush << EnableIds;
  ProcessSkypeMsgs(sc, tm);
  
  Test test(sc, sc.GetLog());
  
  const TestAttr& tst_attr = *_attrs.begin();
  if(DbgMode == debug) {
	BOOST_CHECK_THROW(test[tst_attr], AssertionViolation);
  };
  
  typename Test::const_iterator p = test.find(tst_attr);
  BOOST_CHECK(p == test.end());
  
  test.AskAttr(tst_attr);
  ProcessSkypeMsgs(sc, tm);
  BOOST_CHECK_NO_THROW(r = GetFromString<Result>(test[tst_attr]));
  p = test.find(tst_attr);
  BOOST_CHECK(p == test.begin());
  
  typename std::list<TestAttr>::const_iterator i = _attrs.begin(); 
  for(++i;i != _attrs.end(); ++i) {
	test.AskAttr(*i);
  };
  ProcessSkypeMsgs(sc, tm); 
  std::for_each(test.begin(), test.end(), PrintValue<TestAttr>);
};

void AudioSettingsTest()
{
  SkypeConnection& sc = *SC;

  std::list<AudioAttrs_t> attrs;
  attrs.push_back(auAudioIn);
  attrs.push_back(auAudioOut);
  attrs.push_back(auRinger);
  
  DoTest<AudioSettings, AudioAttrs_t, std::string>(attrs);

  attrs.clear();
  attrs.push_back(auPCSpeaker);
  attrs.push_back(auAGC); 
  attrs.push_back(auAEC);
  attrs.push_back(auMute);
  DoTest<AudioSettings, AudioAttrs_t, bool>(attrs);
};


test_suite* init_unit_test_suite(int argc, char* argv[])
{
    test_suite* test= BOOST_TEST_SUITE("Audio settings unit test");
    const boost::posix_time::seconds tm(1);

    static FileLogger<> fl(stderr, LlDbg);
    LogFunc_t log(&fl, &FileLogger<>::operator());
    static SimpleHandler_t def(DefaultCallback);
    static SkypeConnection scl(argv[0], def, log);
    scl.Attach();
   
    SC = &scl;
    ProcessSkypeMsgs(scl, tm);

    test->add(BOOST_TEST_CASE(&WatchTestAudio));
    test->add(BOOST_TEST_CASE(&AudioSettingsTest));
    return test;
};

