/*!
  \file unix/LowLevel-dbus.cpp

  \brief Low-level ++Skype API.  Unix implementation. D-BUS variant

  $Id $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   18.10.2006 23:29:01 MSD
*/

#include "../h/Cfg.h"

#ifdef USE_DBUS_INTERFACE
#ifdef SKYPE_NIX

#include "../h/LowLevel.h"

UnixDbusRawSkypeConnection_t::RawSkypeConnection(const RawNotify_t& _cbk, const std::string& _app_name, LogFunc_t _log_func) 
			: SkypeCmdFormatter< RawSkypeConnection<os_nix> >(*this),
			  LogAbility(_log_func),
			  cbk(_cbk), app_name(_app_name), dbus_watch(0), attach_result(false), pid(getpid())
{ dbus_error_init(&dbus_error); };	

void UnixDbusRawSkypeConnection_t::Attach(DBusType_t _bus_tp, bool _dont_select, unsigned _timeout)
{
  if(attach_result) return;

  dont_select	= _dont_select;
  timeout	= _timeout;

  DBusBusType tp = (_bus_tp == SessionBus) ? DBUS_BUS_SESSION : DBUS_BUS_SYSTEM;
  dbus_cn = dbus_bus_get(tp, &dbus_error); 
  CheckError();

  // Is Skype registered on the bus?
  DBusMessage* msg = dbus_message_new_method_call("org.freedesktop.DBus", "/org/freedesktop/DBus", 
						  "org.freedesktop.DBus", "ServiceExists");
  dbus_msg_add_arg(msg, "com.Skype.API");
  DBusMessage *reply = SendMsgWBlock(*msg);
  InitDbusIter(*reply);  
  bool r = GetArgBool(dbus_iter);
  if(!r) {
	dbus_message_unref(reply);
	throw SkypeDisconnect("Skype is not registered on the D-BUS"); 
  };		  
  dbus_message_unref(reply);
  
  // Pinging Skype with "Ping" command.  Simply check that service can reponse to "Ping" command
  msg =  dbus_message_new_method_call("com.Skype.API", "/com/Skype", "com.Skype.API", "Ping");
  reply = SendMsgWBlock(*msg);
  dbus_message_unref(reply);

  /// Am I connected and authentificated?
  if(!dbus_connection_get_is_connected(dbus_cn)) 
	  throw SkypeDisconnect("D-BUS connection to Skype was not established");
  if(!dbus_connection_get_is_authenticated(dbus_cn))
	  throw SkypeDisconnect("D-BUS connection to Skype is not authentificated");
  
  // Sending "NAME" command to Skype - some kind of app <-> skype handshaking
  msg =  dbus_message_new_method_call("com.Skype.API", "/com/Skype", "com.Skype.API", "Invoke");
  std::string cmd = "NAME "; cmd.append(app_name);
  dbus_msg_add_arg(msg, cmd);
  reply = SendMsgWBlock(*msg);
  InitDbusIter(*reply);  
  std::string s = GetArgStr(dbus_iter);
  if(s.compare("OK")) {
	dbus_message_unref(reply);
	throw SkypeAccessDenied();
  };		  
  dbus_message_unref(reply);
 
  // Registering interface to get notifications and cmd responses from Skype
  dbus_connection_set_watch_functions(dbus_cn, dbus_add_watch, dbus_remove_watch, NULL, this, NULL);

  DBusObjectPathVTable vtable = { &dbus_unregister_handler, &dbus_msg_handler, NULL, NULL, NULL, NULL };
  bool success = dbus_connection_register_object_path(dbus_cn, "/com/Skype/Client", &vtable, this);
  if(!success) throw SkypeDisconnect("Couldn't register object '/com/Skype/Client' on D-BUS.  Not enough memory?"); 

  attach_result = true;
};

void UnixDbusRawSkypeConnection_t::Detach()
{
  dbus_error_free(&dbus_error);	
};

void UnixDbusRawSkypeConnection_t::flush(const std::string& _cmd)
{
  log(LlDbg, "%s: Sending to SKYPE: '%s'", __PRETTY_FUNCTION__, _cmd.c_str());	
  
  DBusMessage *msg =  dbus_message_new_method_call("com.Skype.API", "/com/Skype", "com.Skype.API", "Invoke");
  dbus_msg_add_arg(msg, _cmd);
  DBusMessage *reply = SendMsgWBlock(*msg);
  dbus_message_unref(msg);
  dbus_replies.push_back(reply);
};

void UnixDbusRawSkypeConnection_t::flush()
{
	SkypeCmdFormatter< RawSkypeConnection<os_nix> >::flush();
}

bool UnixDbusRawSkypeConnection_t::operator()()
{
  Assert(attach_result);

  int retval;
  bool cached_replies = !dbus_replies.empty();

  if(cached_replies) while(!dbus_replies.empty()) {
				DBusMessage*& r = dbus_replies.front();  
				InitDbusIter(*r);
				std::string s = GetArgStr(dbus_iter);
				dbus_message_unref(r);
				dbus_replies.pop_front();
				cbk(s);
  		     };
  
  if(!dont_select) {
	struct timeval tv;
	tv.tv_sec = timeout / 1000000; tv.tv_usec = timeout % 1000000;
	FD_SET(dbus_fd, &read_fds);
  	retval = select(dbus_fd + 1, &read_fds, NULL, NULL, &tv);
  } else retval = 1;

  if(retval) {
	dbus_watch_handle(dbus_watch, DBUS_WATCH_READABLE);
	while (dbus_connection_dispatch(dbus_cn) == DBUS_DISPATCH_DATA_REMAINS) ;
  };

  if(cached_replies) retval = 1;
  
  return retval;
};

DBusHandlerResult UnixDbusRawSkypeConnection_t::dbus_msg_handler(DBusConnection *_connection, DBusMessage *_msg, void *_user_data)
{
  const char* method = dbus_message_get_member(_msg);
 
  if(!method) return DBUS_HANDLER_RESULT_HANDLED;

  // _user_data stores the pointer to UnixDbusRawSkypeConnection_t instance
  Assert(_user_data);
  UnixDbusRawSkypeConnection_t *obj = (UnixDbusRawSkypeConnection_t*) _user_data;

  // First, we should send reply to the Skype if it is requested by Skype
  if(!dbus_message_get_no_reply(_msg)) {
	DBusMessage* m = dbus_message_new_method_return(_msg);
	Assert(m != NULL);
	if(std::string(method).compare("Ping") == 0) { // Ping was weceived.  Answering with PID
		obj->dbus_msg_add_arg(m, obj->pid);
	} else { // Notify was received (or something else, it doesn't matter?)
		obj->dbus_msg_add_arg(m, "ERROR 2");
	};
	dbus_bool_t r = dbus_connection_send(obj->dbus_cn, m, 0);
	Assert(r == TRUE); 
	dbus_connection_flush(obj->dbus_cn);
	dbus_message_unref(m);
  };
  
  // Incoming message handling
  obj->InitDbusIter(*_msg);
  std::string s = obj->GetArgStr(obj->dbus_iter);
  obj->cbk(s);

  return DBUS_HANDLER_RESULT_HANDLED;
};

dbus_bool_t UnixDbusRawSkypeConnection_t::dbus_add_watch(DBusWatch* _watch, void* _data)
{
  UnixDbusRawSkypeConnection_t* obj = (UnixDbusRawSkypeConnection_t*) _data;
  FD_ZERO(&obj->read_fds);
  if(dbus_watch_get_enabled(_watch)) {
	obj->dbus_watch = _watch;
	obj->dbus_fd = dbus_watch_get_fd(_watch);
	Assert(dbus_watch_get_flags(_watch) == DBUS_WATCH_READABLE);
  } else obj->dbus_watch = 0;
  
  return true;
};

#endif // USE_DBUS_INTERFACE
#endif // SKYPE_NIX 

